// ctrlr8 - ATtiny85 timing using timer1 interrupt

// Started: Feb 27, 2015  G. D. (Joe) Young <jyoung@islandnet.com>
//
// Revised: Mar/Apr/15 - see codppmx7 for earlier version(s)
//          Sep 13/15 - codppmx8 adds callsign identification to all modes. From condev2
//          Sep 16/15 - set callspeed to 40 wpm, same format all modes, detect no callsign
//          Jan 11/16 - invert Tx output for FET keying of Tx
//          Feb 27/16 - fix LED-on error, add invert possibility for LED as well as for Tx
//          Jan 04/17 - codppmx8les - add compile-time option to index start in mode 3
//          Mar 03/17 - ctrlr8 - testing more modes via pin 7 A/D
//          Mar 04/17 - re-assign modes
//          Mar 05/17 - cleanup, default not used
//          May 12/17 - retrofit developments from ctrlr14; ctrlr8.h mainly
//          May 17/17 - reduced overlap and no ovelap options
//          May 18/17 - SSS mode for sprint centre beacon
//          May 26/17 - bring adjOverlap( ) macro from ctrlr14.h
//
// ctrlr8 will eventually include all the compile-time options that have been tested
// in other versions of codppx.... In addition, the modification to incorporate fox ID
// on the FoxOr continuous mode is included which requires more modes be available
//
//
// To permit re-programming the callsign without otherwise modifying the controller
// code, the fox ID (MOE..MO5) and the call are stored in processor's eeprom at 
// location 0, in the form of morse codewords (see codebytes.txt).

// An example eeprom entry:
// >>> dump eeprom 
// 0000  07 0f 08 00 0c 02 00 11  02 38 18 12 0d 32 38 ff  |.... ....8.. 28.|
// which corresponds to the text: MOS de VE7BFK/7

// The codeword storage format is used to avoid storing the table of ASCII-to-morse
// An arduino utility, text2code, can be used to produce an intel-hex file for loading
// into eeprom.



// There are 8 modes of operation:
//  mode  operation of   typical fmode
//  name   mode           assignment (see ctrlr8.h for actual)
//   FOS  FoxOr slow            0
//   SPS  Sprint slow           1
//   BEA  continuous MO         2
//   NOR  normal 1min/4min      3
//   NU1  not used              4
//   SSS  sprint centre beacon  5
//   SPF  Sprint fast           6
//   FOF  FoxOr fast            7



#include <avr/io.h>
#include <avr/interrupt.h>
#include <EEPROM.h>

// ***** compile-time options
//#define INDEXMODENOR         // calculate start time from fox number in normal mode

#define TPOLARITY_OD       // Tx output open-drain negative true
//#define TPOLARITY_POS      // Tx output active, positive true
//#define TPOLARITY_NEG       // Tx output active, negative true
#define LPOLARITY_OD        // LED output open-drain negative true
//#define LPOLARITY_POS      // LED output active, positive true

#define CALL_INTERVAL 1800  // time interval between sending callsign

#define NRMODEBITS 3        // bits used for mode selection: 3, 2, or 1
//#define ALTMODES1BIT2       // enable alternative assignment of 4 modes

//#define TEST                // produce test o/p on select pin

//#define ADJ_OVERLAP 1500  // divisor for calcn of seconds to shorten on time

const byte ledPin = 0;
const byte txPin = 1;
const byte ainPin = A1;

#include "ctrlr8.h"

byte fmode, thisfox;
word onTime, offTime, ledtime, calltime;
boolean code = false;              // true during code sending
boolean sendm, sendo, msent, sendid, sendsp, cend, ledon;
boolean sendlsp, sendwsp, eltOn;   // loop control of morse output
unsigned long timer;
unsigned long dotLen = 221;        // dot length in millisec (1200/wpm)1.8432
unsigned long cdotLen = 55;  // dot length while sending call sign
unsigned long savedotLen;

// interrupt service variables
volatile word ovflcnt = 0;
word loopovfl;
bool testtog;

// change interrupt service to compare match interrupt
ISR( TIM1_COMPA_vect ) {
  ovflcnt += 1;
#ifdef TEST
  testtog = !testtog;
  digitalWrite( 2, testtog );
#endif
}

void setupTimer1( ) {
  TCCR1 = bit(CTC1) | 0x0e; // do not use pwm mode
  OCR1C = 224;    // 225 div with 1.8432 MHz xtal
  OCR1A = 224;    // using compare match int
  TIMSK |= bit(OCIE1A);  //add timer1 overflow interrupt
} // setupTimer1 for interrupt generation on overflow

// new functions and variables for callsign insertion
void setupNext( byte );       // functions to setup pointers to code bytes
void setupCall( );
byte setGetBit( );    // code byte parsing, now returns fence bit position 7..0
bool getBit( );
byte codeChar, nextelt, nextBit, bitCnt, nextChar, thisLength;
byte callend;
bool endCode, callsend;



void setup( ) {

  int temp = analogRead( ainPin );      // get operating mode
  fmode = temp>>7;                      // Mar 3/17 - 8 modes
  if( NRMODEBITS < 3 ) fmode = fmode>>1;
  if( NRMODEBITS == 1 ) fmode = fmode>>1;
  if( fmode==NU1 ) fmode = FOS;         // default to continuous for not used
//  if( fmode==NU2 ) fmode = FOF;   // now use for SSS

  onTime = 12;
  offTime = 48;
  if( fmode == SPF || fmode == FOF ) dotLen -= 64;        // fast sprint, FoxOr
  savedotLen = dotLen;
  if( fmode == NOR ) {
    onTime = 60;
    offTime = 240;
  } // if normal timing
  noInterrupts( );    // attempt to sync millis( ) with timer1
  setupTimer1( );
  interrupts( );
#ifdef TEST
  pinMode( 2, OUTPUT );  // turn pin around for test mode
#endif
  setLEDpin( );
  setTxpin( );
  eltOn = false;
  cend = false;
  ledon = true;

  codeChar = EEPROM.read( 2 );  //determine fox number from EE code character
  thisfox = setGetBit( );
  callend = 4;                  //locate end of callsign
  while( EEPROM.read( callend ) != 0xff ) callend++;

  thisLength = 3;
  if( fmode == BEA || fmode == SSS ) {
    thisfox = 0;              // no fox id for continuous
    thisLength = 2;
    ledtime = ovflcnt + 30;   // 30 sec for led 
  }
  if( fmode==FOF || fmode==FOS ) ledtime = ovflcnt + 30;  //led at start of FoxOr modes

// Setup timers
  if( fmode == SPF || fmode == SPS ) {
    loopovfl = ovflcnt+1+(thisfox-1)*onTime;  //indexed start for sprint
  } else {
    loopovfl = ovflcnt + 1;    // no indexing for normal
  } // if sprint modes

#ifdef INDEXMODENOR
// Also index start time from fox number if in normal mode
  if( fmode == NOR ) {
    loopovfl = ovflcnt+1+(thisfox-1)*onTime;  //indexed start for normal
  } // if mode NOR to be indexed start
#endif

  adjOverlap( );      // use macro if reducing overlap of transmissions
  
  calltime = ovflcnt+2;

} // setup( )


void loop( ) {

  if( !code && loopovfl == ovflcnt ) {  // start sending
    loopovfl += onTime;
//    if( fmode == 3 ) {
//      setupCall( );               // mode 3 send call every start
//    } else {
      dotLen = savedotLen;
      nextChar = 0;
      setupNext( nextChar );      // get first letter of fox code
//    }
    code = true;
    cend = false;
  }

// end sending
  if( code && loopovfl == ovflcnt && (fmode==SPS || fmode==SPF || fmode==NOR) ) {
    loopovfl += offTime;
    cend = true;    // delay end until id complete
  }

  if( (fmode==BEA || fmode==FOF || fmode==FOS || fmode==SSS ) && ledtime == ovflcnt ) ledon = false;
  
  if( (ovflcnt-calltime) > CALL_INTERVAL ) {
    calltime = ovflcnt;
    callsend = true;
  }
  
  if( code && !eltOn ) {
    timer = millis( ) + dotLen;            // setup for dot length
    if( getBit( ) ) timer += 2*dotLen;      // sending a dash
    TxpinON( );
    if( ledon ) LEDpinON( );
    eltOn = true;
  } // if not sending, start first element

  if( code && eltOn && !sendsp && ( timer < millis( ) ) ) {
    TxpinOFF( );
    LEDpinOFF( );
    timer += dotLen;    // timing space following element
    sendsp = true;
    if( endCode ) {
      endCode = false;
      sendlsp = true;
      timer += 2*dotLen;
      nextChar++;
      if( (nextChar == thisLength) || (nextChar == callend) ) {
        dotLen = savedotLen;
        sendwsp = true;
        timer += 3*dotLen;
        nextChar = 0;          // reset pointer to characters
      } // if last character of this group was just sent
      setupNext( nextChar );
      if( codeChar == 0 ) {
        sendwsp = true;
        timer += 3*dotLen;
        nextChar++;
        if( nextChar == thisLength ) nextChar = 0;
        setupNext( nextChar );    // if space charcter to be sent
      }
    } // extend space to letter space or word space
  } // if element on time is up

  if( code && sendsp && ( timer < millis( ) ) ) {
    eltOn = false;
    sendsp = false;
    if( sendlsp ) {
      sendlsp = false;
    } // if letter space just ended
    if( sendwsp ) {
      sendwsp = false;
      if( cend ) {
        code = false;  // turn off sending
        ledon = false;   // and led after first burst
      } // if last word space, end sending
      if( callsend ) {
        callsend = false;
        setupCall( );
      } // if time to send a callsign
    } // if word space ended, setup to begin again
  } // if sending space 
  
} // loop( )


// support function implementations
byte ssschars[ ] = { 0x08, 0x00, 0x00 };
void setupNext( byte nextChar ) {
  if( fmode == SSS && nextChar<4 ) {
    codeChar = ssschars[ nextChar ];
  } else {
    codeChar = EEPROM.read( nextChar );
  }
  setGetBit( );
} // setup to first code byte (usually for M)

void setupCall( ) {
  nextChar =  4;    // start of callsign in EE
  codeChar = EEPROM.read( nextChar );
  dotLen = cdotLen;
  if( codeChar == 0xff ) {              // allow for suppressing call
    nextChar = 0;
    codeChar = EEPROM.read( nextChar );
    dotLen = savedotLen;
  }
  setGetBit( );
} // setup to first code byte of callsign ID

byte setGetBit( ) {
  bitCnt = 7;
  nextBit = 0b10000000;
  while( (bitCnt > 0) && ((nextBit & codeChar) == 0) ) {
    nextBit = nextBit>>1;
    bitCnt--;
  } // while leading zeros before fence bit
  return bitCnt;
} // setup the code word parsing

bool getBit( ) {
  nextBit = nextBit>>1;
  bitCnt--;
  if( bitCnt == 0 ) {
    endCode = true;
  } // if last bit has just been selected
  return( (nextBit & codeChar) == nextBit );
} // get next code element to send


